/************************************************************************/
/* ioloop.c							     */
/************************************************************************/
/*
 *  (c)2021,2022 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdint.h>
#include <sys/epoll.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <ioloop.h>


#define READ_FD			0
#define WRITE_FD		1

#define FD_EMPTY		-1

ioloop_t *ioloop_open(int options)
{
	int			rc;
	int			flags;
	int			i;
	struct epoll_event	epevent;
	ioloop_t		*loop = NULL;

	loop = calloc(1, sizeof(ioloop_t));
	if (!loop)
		return NULL;
	loop->poll_timeout = IOLOOP_DEFAULT_TIMEOUT;

	// invalidate all fds in the table
	for (i=0 ; i<EPOLL_TABLE_SIZE ; i++) {
		loop->fd_descriptor[i] = FD_EMPTY;
	}

	loop->epoll_pipe[READ_FD] = -1;
	loop->epoll_pipe[WRITE_FD] = -1;
	rc = pipe(loop->epoll_pipe);
	if (rc == -1)
		goto ioloop_open_exit;
	flags = fcntl(loop->epoll_pipe[READ_FD], F_GETFL, 0);
	rc = fcntl(loop->epoll_pipe[READ_FD],
			F_SETFL,
			flags | O_NONBLOCK);
	if (rc == -1)
		goto ioloop_open_exit;

	loop->epoll_fd = epoll_create(EPOLL_TABLE_SIZE);
	if (loop->epoll_fd == -1)
		goto ioloop_open_exit;

	epevent.data.fd = loop->epoll_pipe[READ_FD];
	epevent.events = EPOLLIN;
	rc = epoll_ctl(loop->epoll_fd,
			EPOLL_CTL_ADD,
			loop->epoll_pipe[READ_FD],
			&epevent);
	if (rc == -1)
		goto ioloop_open_exit;

	return loop;

ioloop_open_exit:
	if (loop->epoll_fd != -1)
		(void)close(loop->epoll_fd);
	if (loop->epoll_pipe[0] != -1)
		(void)close(loop->epoll_pipe[0]);
	if (loop->epoll_pipe[1] != -1)
		(void)close(loop->epoll_pipe[1]);
	if (loop)
		free(loop);
	return NULL;
}


int ioloop_close(ioloop_t *loop)
{
	if (loop->epoll_fd != -1)
		(void)close(loop->epoll_fd);
	if (loop->epoll_pipe[0] != -1)
		(void)close(loop->epoll_pipe[0]);
	if (loop->epoll_pipe[1] != -1)
		(void)close(loop->epoll_pipe[1]);
	if (loop)
		free(loop);
	return 0;
}


int ioloop_set_default_io_callback(ioloop_t *loop, ioloop_callback_func func)
{
	loop->default_callback = func;
	return 0;
}


int ioloop_register_fd(ioloop_t *loop,
			int fd,
			uint8_t type,
			uint16_t port,
			ioloop_callback_func callback,
			void *callback_arg)
{
	int			rc;
	int			i;
	struct epoll_event	epevent;

	if (loop->fd_count >= POLL_MAX_SOCKETS)
		return ENOSPC;

	// locate empty table slot
	// TODO eliminate this (just hash on the fd)
	for (i=0 ; i<EPOLL_TABLE_SIZE ; i++) {
		if (loop->fd_descriptor[i] == FD_EMPTY)
			break;
	}
	if (i == EPOLL_TABLE_SIZE)
		return ENOSPC;


//	printf("DEBUG %s() fd=%d loop->fd_count=%d callback=%p\n", __func__, fd, loop->fd_count, callback);

	epevent.data.u32 = i;	// store fd array offset
	epevent.events = (EPOLLIN | EPOLLERR | EPOLLHUP);
	rc = epoll_ctl(loop->epoll_fd, EPOLL_CTL_ADD, fd, &epevent);
	if (rc == -1)
		return errno;

	loop->fd_descriptor[i] = fd;
	loop->poll_conn_data[i].type = type;
	loop->poll_conn_data[i].port = port;
	loop->fd_callback[i] = callback;
	loop->fd_callback_arg[i] = callback_arg;
	loop->fd_count++;

	return 0;
}


int ioloop_unregister_fd(ioloop_t *loop, int fd)
{
	int			rc;
	int			i;
	int			oset;

	if (loop->fd_count >= POLL_MAX_SOCKETS)
		return ENOSPC;

	// locate the fd in the array
	// TODO eliminate this (just hash on the fd)
	for (i=0 ; i<loop->fd_count ; i++) {
		if (loop->fd_descriptor[i] == fd)
			break;
	}
	if (i == loop->fd_count)
		return ENOENT;
	oset = i;

	// remove the fd from epoll
	rc = epoll_ctl(loop->epoll_fd, EPOLL_CTL_DEL, fd, NULL);
	if (rc == -1) {
		rc = errno;
		printf("ERROR %s() epoll_ctl() fd=%d rc=%d\n", __func__, fd, rc);
		return rc;
	}

	// clear the entry
	loop->fd_descriptor[oset] = -1;
	loop->poll_conn_data[oset].type = 0;
	loop->poll_conn_data[oset].port = 0;
	loop->fd_callback[oset] = NULL;
	loop->fd_callback_arg[oset] = NULL;
	// update the table entry count
	loop->fd_count--;

	return 0;
}


int ioloop_modify_fd(ioloop_t *loop,
			int fd,
			ioloop_modify_opt_t option,
			ioloop_modify_arg_t *arg)
{
	int			rc;
	int			ti;
	struct epoll_event	epevent;

	// locate table slot for fd
	// TODO eliminate this (just hash on the fd)
	for (ti=0 ; ti<EPOLL_TABLE_SIZE ; ti++) {
		if (loop->fd_descriptor[ti] == fd)
			break;
	}
	if (ti == EPOLL_TABLE_SIZE)
		return ENOENT;

	switch(option) {
		case IOLOOP_MODIFY_FD_O_EVENTS:
			epevent.events = arg->u.events;
			epevent.data.u32 = ti;
			rc = epoll_ctl(loop->epoll_fd,
					EPOLL_CTL_MOD,
					fd,
					&epevent);
			if (rc == -1)
				return errno;
			break;
		default:
			return EINVAL;
	}
	return 0;
}


int ioloop_run(ioloop_t *loop, int options, uint32_t timeout)
{
	int			rc;
	int			i;
	int			efd;
	int			count;
	int			err_count;
	char			c;
	ioloop_callback_func	cb;
	struct epoll_event      epevent[EPOLL_TABLE_SIZE];
	
	err_count = 0;
	while (1) {
		rc = epoll_wait(loop->epoll_fd,
				epevent,
				EPOLL_TABLE_SIZE,
				loop->poll_timeout);
		if (rc == -1)
			return errno;
		if (rc == 0) {
			// handle timeout and continue
			continue;
		}

		count = rc;
		for (i=0 ; i<count ; i++) {

			efd = epevent[i].data.u32;
			// bad fd, shouldn't happen but....
			if (loop->fd_descriptor[efd] == -1) {
				printf("ERROR %s() bad file descriptor"
					" in table fd=%d\n",
					__func__,
					loop->fd_descriptor[efd]);
				err_count++;
				if (err_count >= 5)
					return 2;
			}

			// handle command pipe
			if (epevent[i].events & EPOLLIN &&
			    loop->fd_descriptor[efd] == loop->epoll_pipe[READ_FD]) {
				rc = read (loop->epoll_pipe[READ_FD], &c, 1);
				if (c == 'E') // loop exit triggered
					return 0;
				continue;
			}

			// handle application fds
			cb = NULL;
			if (loop->fd_callback[efd] != NULL)
				cb = loop->fd_callback[efd];
			else if (loop->default_callback != NULL)
				cb = loop->default_callback;
			if (cb) {
				rc = (*cb)(loop->fd_descriptor[efd],
					epevent[i].events,
					&loop->poll_conn_data[efd],
					loop->fd_callback_arg[efd]);
				(void)rc;
			} else {
				printf("ERROR %s() NULL callback for fd=%d\n",
					__func__,
					loop->fd_descriptor[efd]);
				err_count++;
				if (err_count >= 5)
					return 2;
			}
		}
	}
	return 0;
}


int ioloop_exit(ioloop_t *loop)
{
	int	rc;
	rc = write (loop->epoll_pipe[WRITE_FD], "E", 1);
	return rc;
}


