/************************************************************************/
/* pfm_milter.c - SMTP Milter filter module				*/
/************************************************************************/
/*
 *  (c)2021,2022 Todd Shadburn <tshadburn@opencomputingsolutions.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/************************************************************************/


#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <cutil/cutil.h>
#include <pcre.h>

#include <pnetbuffer.h>
#include <logging.h>
#include <tcp.h>
#include <pfm.h>
#include <metric.h>
#include <microfsm.h>
#include <helpers.h>

#include <milter.h>
#include <filterset.h>


#define PACKET_BUFFER_SIZE		64512
#define PFM_MILTER_IP_DEFAULT		"0.0.0.0"
#define PFM_MILTER_PORT_DEFAULT		5025
#define RESPONSE_ARRAY_SIZE		4
#define RESPONSE_STRING_SIZE		256
#define MAIL_CONTEXT_STRING_SIZE	256
#define LINEBUFFER_SIZE			2048

#define VOTE_PASS			0
#define VOTE_REJECT			1
#define VOTE_REMOVE			2
#define VOTE_LOG			3


enum milter_signals {
	SIG_MILTER_REQUEST		= MICROFSM_SIG_USER,
	SIG_MILTER_RESET,
	SIG_MILTER_RAW_DATA,
};


typedef struct _milter_event_t milter_event_t;
struct _milter_event_t {
	microfsm_event_t	super;

	pnetbuffer_t		*nb_in;
	pnetbuffer_t		*nb_out;
};

#define CONN_TABLE_SIZE_DEFAULT		256

const char *vote_string[] = {
	"pass",
	"reject",
	"remove",
	"log",
};


typedef enum {
	METRIC_CONN_TOTAL	= 0,
	METRIC_CONN_CURRENT,
	METRIC_CONN_CLIENT_DROP,
	METRIC_CONN_SERVER_DROP,
	METRIC_CONN_PROTOCOL_VIOLATION,
	METRIC_MAIL_TOTAL,
	METRIC_MAIL_PASS,
	METRIC_MAIL_REJECT,
	_METRIC_COUNT,
} milter_metric_index;


enum milter_state_index {
	MILTER_STATE_CONNECT		= 0,
	MILTER_STATE_HELO,
	MILTER_STATE_MAILFROM,
	MILTER_STATE_RCPTTO,
	MILTER_STATE_HEADER,
	MILTER_STATE_BODY,
	_MILTER_STATE_COUNT,
};


char *milter_state_name[] = {
	"connect",
	"helo",
	"mail",
	"rcpt",
	"header",
	"body",
};


typedef struct _mail_context_t mail_context_t;
struct _mail_context_t {
	char		conn_hostname[MAIL_CONTEXT_STRING_SIZE];
	char		conn_addr[MAIL_CONTEXT_STRING_SIZE];
	char		helo[MAIL_CONTEXT_STRING_SIZE];
	char		mailfrom[MAIL_CONTEXT_STRING_SIZE];
	char		rcptto[MAIL_CONTEXT_STRING_SIZE];

	int32_t		score;
};


typedef enum {
	CONN_IO_S_READ_PROTO	= 0,
	CONN_IO_S_READ_RAW,
	CONN_IO_S_WRITE,
	_CONN_IO_S_MAX,
} conn_io_state;


typedef struct _milter_conn_t milter_conn_t;
struct _milter_conn_t {

	int		fd;

	microfsm_t	fsm;

	pnetbuffer_t		*nb_in;
	pnetbuffer_t		*nb_out;
	char			*buffer_in;
	char			*buffer_out;

	mail_context_t		ctx;

	conn_io_state		io_state;
	uint32_t		rem_read_bytes;

	uint32_t		body_total_len;
};



typedef struct _pfm_milter_t pfm_milter_t;
struct _pfm_milter_t {
	pfm_module_t	*module;	// passed in module_open()

	uint64_t	policy_bits;

	char		**command_array;
	char		**response_array;

	metric_t	metric[_METRIC_COUNT];

	SList		*q_filter_list;
	SList		*an_filter_list;

	char		*service_bind_address;

	int		conn_table_size;
	milter_conn_t	*conn_table;

	// all defined filtersets
	SList		*filterset_list;

	// filterset state attachments
	SList		*state_filterset_list[_MILTER_STATE_COUNT];

	// policy elements
	int32_t		reject_score;	// email scores >= this
					// will be rejected

};


// policy bits management
typedef enum {
	POLICY_NOOP				= 0,
	POLICY_DENY_MULTIPLE_QUERIES,

	_POLICY_MAXIMUM,
} policy_bit_t;

typedef struct _policy_item_t policy_item_t;
struct _policy_item_t {
	char		*name;
	policy_bit_t	bit; // policy_bits |= (1<<bit)
};

policy_item_t policy_table[] = {
	{"deny-multiquery",	POLICY_DENY_MULTIPLE_QUERIES},
	{NULL,			POLICY_NOOP},
};


typedef struct _filterset_enum_arg_t filterset_enum_arg_t;
struct _filterset_enum_arg_t {
	filterset_t		*fs;
	pnetbuffer_t		*nb;
};



// Forward prototypes
int vector_process_command(pfm_t *pfm,
			pnetbuffer_t *indata,
			pnetbuffer_t *outdata);
int vector_handle_new_connection(pfm_t *pfm,
			io_context_t *ctx,
			pnetbuffer_t *indata,
			pnetbuffer_t *outdata);
int milter_io_callback(int fd, int events, connection_data_t *cd, void *arg);
int server_io_callback(int fd, int events, connection_data_t *cd, void *arg);

microfsm_result_t milter_state_optneg(microfsm_t *me, microfsm_event_t *event);
microfsm_result_t milter_state_connect(microfsm_t *me, microfsm_event_t *event);
microfsm_result_t milter_state_hello(microfsm_t *me, microfsm_event_t *event);
microfsm_result_t milter_state_mailfrom(microfsm_t *me, microfsm_event_t *event);
microfsm_result_t milter_state_rcptto(microfsm_t *me, microfsm_event_t *event);
microfsm_result_t milter_state_headers(microfsm_t *me, microfsm_event_t *event);
microfsm_result_t milter_state_body(microfsm_t *me, microfsm_event_t *event);
int milter_dump_config(pnetbuffer_t *nb_out);


// Module Globals
pfm_vectors_t vectors = {
	.process_command	= vector_process_command,
	.process_data		= vector_handle_new_connection,
};

uint16_t port_list[] = {
	PFM_MILTER_PORT_DEFAULT,
	0,
};

pfm_t pfm_info = {
	.version	= PFM_VERSION_1,
	.flags		= 0,
	.name		= "milter",
	.vectors	= &vectors,
	.userdata	= NULL,
};


pfm_milter_t	priv_data;



int pfm_module_info(pfm_t *pfm)
{
	pfm->version = pfm_info.version;
	pfm->flags = pfm_info.flags;
	pfm->name = pfm_info.name;
	pfm->vectors = pfm_info.vectors;
	pfm->userdata = pfm_info.userdata;
	return 0;
}


int pfm_module_open(pfm_module_t *module)
{
	int			rc;
	int			i;
	int			fd;
	ioctl_opt_register_fd_t opt;
	ioctl_opt_get_ci_t	o_gci;
	char			cfgstr[RESPONSE_STRING_SIZE];

	memset(&priv_data, 0, sizeof(pfm_milter_t));
	priv_data.module = (pfm_module_t *)module;
	priv_data.service_bind_address = PFM_MILTER_IP_DEFAULT;

	// assign default port(s)
	pfm_info.port_binds = port_list;

	// allocate connection table
	priv_data.conn_table_size = CONN_TABLE_SIZE_DEFAULT;
	priv_data.conn_table = calloc(priv_data.conn_table_size, sizeof(milter_conn_t));
	if (!priv_data.conn_table) {
		rc = errno;
		goto pfm_module_open_exit;
	}
	for (i=0 ; i<priv_data.conn_table_size ; i++) {
		priv_data.conn_table[i].buffer_in = malloc(PACKET_BUFFER_SIZE);
		priv_data.conn_table[i].buffer_out = malloc(PACKET_BUFFER_SIZE);
		if ((!priv_data.conn_table[i].buffer_in) ||
		    (!priv_data.conn_table[i].buffer_out)) {
			rc = ENOMEM;
			goto pfm_module_open_exit;
		}
		priv_data.conn_table[i].nb_in = pnetbuffer_init(
					priv_data.conn_table[i].buffer_in,
					PACKET_BUFFER_SIZE);
		priv_data.conn_table[i].nb_out = pnetbuffer_init(
					priv_data.conn_table[i].buffer_out,
					PACKET_BUFFER_SIZE);
	}

	// allocate command buffer array
	priv_data.command_array = calloc(PFM_COMMAND_TOKEN_MAX, sizeof(char*));
	if (!priv_data.command_array) {
		rc = errno;
		goto pfm_module_open_exit;
	}
	for (i=0 ; i<PFM_COMMAND_TOKEN_MAX ; i++) {
		priv_data.command_array[i] = calloc(1, PFM_COMMAND_STR_SIZE);
		if (!priv_data.command_array[i]) {
			rc = errno;
			goto pfm_module_open_exit;
		}
	}

	// allocate response buffer array
	priv_data.response_array = calloc(RESPONSE_ARRAY_SIZE, sizeof(char*));
	if (!priv_data.response_array) {
		rc = errno;
		goto pfm_module_open_exit;
	}
	for (i=0 ; i<RESPONSE_ARRAY_SIZE ; i++) {
		priv_data.response_array[i] = calloc(1, RESPONSE_STRING_SIZE);
		if (!priv_data.response_array[i]) {
			rc = errno;
			goto pfm_module_open_exit;
		}
	}

	// setup metric table
	metric_init(&priv_data.metric[METRIC_CONN_TOTAL],
		"milter.connections.total",
		METRIC_TYPE_UINT32);
	metric_init(&priv_data.metric[METRIC_CONN_CURRENT],
		"milter.connections.current",
		METRIC_TYPE_UINT32);
	metric_init(&priv_data.metric[METRIC_CONN_CLIENT_DROP],
		"milter.connections.dropped.client",
		METRIC_TYPE_UINT32);
	metric_init(&priv_data.metric[METRIC_CONN_SERVER_DROP],
		"milter.connections.dropped.server",
		METRIC_TYPE_UINT32);
	metric_init(&priv_data.metric[METRIC_CONN_PROTOCOL_VIOLATION],
		"milter.connections.protocol.violation",
		METRIC_TYPE_UINT32);
	metric_init(&priv_data.metric[METRIC_MAIL_TOTAL],
		"milter.mail.total",
		METRIC_TYPE_UINT32);
	metric_init(&priv_data.metric[METRIC_MAIL_PASS],
		"milter.mail.pass",
		METRIC_TYPE_UINT32);
	metric_init(&priv_data.metric[METRIC_MAIL_REJECT],
		"milter.mail.reject",
		METRIC_TYPE_UINT32);

	// get service bind address
	o_gci.ci_id = PFM_CI_BIND_ADDRESS;
	o_gci.value.str = NULL;
	rc = pfm_vector_ioctl(priv_data.module,
		PFM_IOCTL_GET_CONFIG_ITEM,
		&o_gci);
	if (rc) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"Failed to retrieve service bind address"
			" rc=%d",
			rc);
		return rc;
	}
	priv_data.service_bind_address = strdup(o_gci.value.str);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_INFO,
		"Retrieved service bind address is %s",
		priv_data.service_bind_address);

	// create listener sockets and connect to ioloop
	i = 0;
	while (1) {
		if (port_list[i] == 0) break;
		rc = snprintf(cfgstr,
			RESPONSE_STRING_SIZE,
			"%d",
			port_list[i]);
		fd = tcp_bind(priv_data.service_bind_address, cfgstr);
		if (fd == -1) {
			rc = errno;
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"Failed to create server socket for"
				" port=%d rc=%d",
				port_list[i],
				rc);
			return rc;
		}
		opt.fd = fd;
		opt.port = port_list[i];
		opt.callback = server_io_callback;
		opt.callback_arg = NULL;
		rc = pfm_vector_ioctl(priv_data.module,
			PFM_IOCTL_REGISTER_BACKEND_FD,
			&opt);
		if (rc) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_ERROR,
				"Failed to register server FD with ioloop"
				" subsystem rc=%d",
				rc);
			i = tcp_shutdown(fd);
			return rc;
		}
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_NOTICE,
			"Bound new server socket fd=%d port=%d",
			fd,
			port_list[i]);
		i++;
	}


	return 0;

pfm_module_open_exit:
	for (i=0 ; i<priv_data.conn_table_size ; i++) {
		if (priv_data.conn_table[i].buffer_in)
			free(priv_data.conn_table[i].buffer_in);
		if (priv_data.conn_table[i].buffer_out)
			free(priv_data.conn_table[i].buffer_out);
	}
	if (priv_data.conn_table)
		free(priv_data.conn_table);
	if (priv_data.command_array)
		free(priv_data.command_array);
	if (priv_data.response_array)
		free(priv_data.response_array);
	return rc;
}


int pfm_module_close(void *args)
{
	if (priv_data.command_array)
		free(priv_data.command_array);
	return 0;
}



policy_item_t *policy_item_lookup_by_name(char *name)
{
	int		i = 0;
	while (policy_table[i].name != NULL) {
		if (!strcmp(name, policy_table[i].name))
			return &policy_table[i];
		i++;
	}
	return NULL;
}


filterset_t *slist_get_filterset_by_name(SList *list, char *name)
{
	filterset_t	*fs;
	SList		*item;
	item = list;
	while (item) {
		fs = slist_get_data(item);
		if (!strcmp(fs->name, name))
			return fs;
		item = slist_get_next(item);
	}
	return NULL;
}


int milter_is_filterset_attached(filterset_t *fs)
{
	int		i;
	filterset_t	*lfs;
	SList		*item;

	for (i=0 ; i<_MILTER_STATE_COUNT ; i++) {
		item = priv_data.state_filterset_list[i];
		while (item) {
			lfs = slist_get_data(item);
			if (lfs == fs)
				return 1;
			item = slist_get_next(item);
		}
	}
	return 0;
}


int output_filterset_entry(filter_t *f, void *arg)
{
	int		rc;
	pnetbuffer_t	*nb;
	char		fmtstr[256];
	nb = arg;
	rc = snprintf(fmtstr,
			256,
			"%8d %s\n",
			metric_get_u32(&f->hits),
			f->match);
	rc = pnetbuffer_add_string(nb, fmtstr);
	return rc;
}


int vector_process_command(pfm_t *pfm, pnetbuffer_t *indata, pnetbuffer_t *outdata)
{
	int		rc;
	int		i;
	int		type;
	int		state;
	int		action;
	uint8_t		argc;
	int32_t		score;
	char		**argv;
	char		*str1;
	char		*str2;
	filterset_t	*fs;
	SList		*li;
	char		fmtstr[256];

//	printf("%s() entry\n", __func__);
	rc = pnetbuffer_read_strings(indata,
				priv_data.command_array,
				PFM_COMMAND_STR_SIZE,
				PFM_COMMAND_TOKEN_MAX,
				&argc,
				PNETBUFFER_OPT_STRINGS_COUNT_FIRST);
	if (rc != 0) {
		//(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
		(void)pnetbuffer_add_string(outdata, "1 Missing arguments\n");
		return rc;
	}
	if (argc < 1) {
		//(void)pfm_pnetbuffer_add_rc_string(outdata, PFM_ERR_CMD_INVALID);
		(void)pnetbuffer_add_string(outdata, "1 Too few arguments\n");
		return 0;
	}
	argv = priv_data.command_array;
	i = 0;
	score = 0;

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() argv[i]='%s'", __func__, argv[i]);
	if (!strcmp(argv[i], "counters")) {
		i++;
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"%s() argv[i]='%s'", __func__, argv[i]);
		if (!strcmp(argv[i], "show")) {
			for (i=0 ; i<_METRIC_COUNT ; i++) {
				snprintf (priv_data.response_array[0],
					RESPONSE_STRING_SIZE,
					"%s %d\n",
					metric_get_name(&priv_data.metric[i]),
					metric_get_u32(&priv_data.metric[i])
				);
				rc = pnetbuffer_add_string(outdata,
						priv_data.response_array[0]);
				if (rc != 0) {
					pnetbuffer_reset(outdata);
					(void)pnetbuffer_add_string(outdata,
						"102 PFM module buffer too short\n");
					return 0;
				}
			}

		} else if (!strcmp(argv[i], "reset")) {
			for (i=0 ; i<_METRIC_COUNT ; i++) {
				metric_set_u32(&priv_data.metric[i], 0);
			}
			(void)pnetbuffer_add_string(outdata,
				"0 PFM counters reset\n");
		} else {
			(void)pnetbuffer_add_string(outdata,
				"2 Invalid module command\n");
			return 0;
		}
		return 0;

	} else if (!strcmp(argv[i], "dump")) {
		i++;
		if (!strcmp(argv[i], "config")) {
			rc = milter_dump_config(outdata);
			return rc;
		}

	} else if (!strcmp(argv[i], "policy")) {
		i++;
		if (!strcmp(argv[i], "set")) {
			i++;
			if ((argc-i) < 2) {
				(void)pnetbuffer_add_string(outdata,
					"2 Not enough arguments for command\n");
				return 0;
			}
			if (!strcmp(argv[i], "reject_score")) {
				i++;
				priv_data.reject_score = atoi(argv[i]);
				(void)pnetbuffer_add_string(outdata,
					"0 Successfully set reject_score \n");
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_NOTICE,
					"Successfully set reject_score to %d",
					priv_data.reject_score);
				return 0;
			} else {
				(void)pnetbuffer_add_string(outdata,
					"2 Invalid argument for policy set command\n");
				return 0;
			}
		} else {
			(void)pnetbuffer_add_string(outdata,
				"2 Invalid argument for policy command\n");
			return 0;
		}

	} else if (!strcmp(argv[i], "filterset")) {
		i++;
		if (!strcmp(argv[i], "create")) {
			if ((argc-i) < 7) {
				(void)pnetbuffer_add_string(outdata,
					"2 Not enough arguments for command\n");
				return 0;
			}
			// name
			i++;
			if (strcmp(argv[i], "name")) {
				(void)pnetbuffer_add_string(outdata,
					"2 Missing name argument\n");
				return 0;
			}
			i++;
			str1 = argv[i];
			fs = slist_get_filterset_by_name(priv_data.filterset_list,
					str1);
			if (fs) {
				(void)pnetbuffer_add_string(outdata,
					"5 A filterset by that name already exists\n");
				return 0;
			}
			// type
			i++;
			if (strcmp(argv[i], "type")) {
				(void)pnetbuffer_add_string(outdata,
					"2 Missing type argument\n");
				return 0;
			}
			i++;
			type = filterset_string_to_type(argv[i]);
			if (!type) {
				(void)pnetbuffer_add_string(outdata,
					"3 Invalid filterset type\n");
				return 0;
			}
			// default action
			i++;
			if (strcmp(argv[i], "action")) {
				(void)pnetbuffer_add_string(outdata,
					"2 Missing action argument\n");
				return 0;
			}
			i++;
			action = filterset_string_to_action(argv[i]);
			if (!type) {
				(void)pnetbuffer_add_string(outdata,
					"3 Invalid filterset action\n");
				return 0;
			}
			fs = filterset_new(str1, type, action);
			if (!fs) {
				(void)pnetbuffer_add_string(outdata,
					"4 Internal memory error\n");
				return 0;
			}
			rc = slist_item_add(&priv_data.filterset_list, fs);
			if (rc) {
				(void)pnetbuffer_add_string(outdata,
					"4 Failed to add filterset\n");
				return 0;
			}
			(void)pnetbuffer_add_string(outdata,
				"0 Successfully added filterset\n");
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_NOTICE,
				"Added filterset %s", fs->name);
			return 0;

		} else if (!strcmp(argv[i], "entry")) {
			if ((argc-i) < 5) {
				(void)pnetbuffer_add_string(outdata,
					"2 Not enough arguments for command\n");
				return 0;
			}
			// operation
			i++;
			str1 = argv[i];
			if (strcmp(str1, "add") && strcmp(str1, "remove")) {
				(void)pnetbuffer_add_string(outdata,
					"2 Invalid operation argument specified\n");
				return 0;
			}
			// name
			i++;
			if (strcmp(argv[i], "name")) {
				(void)pnetbuffer_add_string(outdata,
					"2 Missing name argument\n");
				return 0;
			}
			i++;
			fs = slist_get_filterset_by_name(priv_data.filterset_list,
					argv[i]);
			if (!fs) {
				(void)pnetbuffer_add_string(outdata,
					"5 A filterset by that name does not exist\n");
				return 0;
			}
			// match
			i++;
			if (strcmp(argv[i], "match")) {
				(void)pnetbuffer_add_string(outdata,
					"2 Missing match keyword\n");
				return 0;
			}
			i++;
			if (i == argc) {
				(void)pnetbuffer_add_string(outdata,
					"2 Missing match argument\n");
				return 0;
			}
			str2 = argv[i];
			i++;
			if (!strcmp(argv[i], "score")) {
				// optional score argument/value
				i++;
				if (i == argc) {
					(void)pnetbuffer_add_string(outdata,
						"2 Missing score argument\n");
					return 0;
				}
				score = atoi(argv[i]);
			}
			if (!strcmp(str1, "add")) {
				rc = filterset_add_entry(fs, str2, score);
				if (rc) {
					(void)pnetbuffer_add_string(outdata,
						"6 Failed to add match filter\n");
					return 0;
				}
				(void)pnetbuffer_add_string(outdata,
					"0 Successfully added match filter\n");
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_NOTICE,
					"Added filterset %s entry: %s score: %d",
					fs->name,
					str2,
					score);
			} else if (!strcmp(str1, "remove")) {
				rc = filterset_remove_entry(fs, str2);
				if (rc) {
					(void)pnetbuffer_add_string(outdata,
						"6 Failed to remove match filter\n");
					return 0;
				}
				(void)pnetbuffer_add_string(outdata,
					"0 Successfully removed match filter\n");
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_NOTICE,
					"Removed filterset %s entry: %s",
					fs->name,
					str2);
			}

		} else if (!strcmp(argv[i], "remove")) {
			if ((argc-i) < 2) {
				(void)pnetbuffer_add_string(outdata,
					"2 Not enough arguments for command\n");
				return 0;
			}
			// name
			i++;
			if (strcmp(argv[i], "name")) {
				(void)pnetbuffer_add_string(outdata,
					"2 Missing name argument\n");
				return 0;
			}
			i++;
			fs = slist_get_filterset_by_name(priv_data.filterset_list,
					argv[i]);
			if (!fs) {
				(void)pnetbuffer_add_string(outdata,
					"5 A filterset by that name does not exist\n");
				return 0;
			}
			if (milter_is_filterset_attached(fs)) {
				(void)pnetbuffer_add_string(outdata,
					"6 That filterset is attached and"
					" cannot be removed\n");
				return 0;
			}
			rc = slist_item_remove(&priv_data.filterset_list, fs);
			if (rc) {
				(void)pnetbuffer_add_string(outdata,
					"4 Failed to remove filterset\n");
				return 0;
			}
			rc = filterset_destroy(fs);
			if (rc) {
				(void)pnetbuffer_add_string(outdata,
					"4 Failed to destroy filterset\n");
				return 0;
			}
			(void)pnetbuffer_add_string(outdata,
				"0 Successfully removed filterset\n");
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_NOTICE,
				"Removed filterset %s", argv[i]);

		} else if (!strcmp(argv[i], "show")) {
			if ((argc-i) < 2) {
				(void)pnetbuffer_add_string(outdata,
					"2 Not enough arguments for command\n");
				return 0;
			}
			// name
			i++;
			if (strcmp(argv[i], "name")) {
				(void)pnetbuffer_add_string(outdata,
					"2 Missing name argument\n");
				return 0;
			}
			i++;
			fs = slist_get_filterset_by_name(priv_data.filterset_list,
					argv[i]);
			if (!fs) {
				(void)pnetbuffer_add_string(outdata,
					"5 A filterset by that name does not exist\n");
				return 0;
			}
			rc = snprintf(fmtstr,
					256,
					"%8s %s\n",
					"Hits",
					"Match");
			(void)pnetbuffer_add_string(outdata, fmtstr);
			rc = filterset_enumerate_entries(fs,
				output_filterset_entry,
				outdata);

		} else if (!strcmp(argv[i], "list")) {
			rc = snprintf(fmtstr,
				256,
				"%-32s %-12s %-12s\n",
				"Name",
				"Type",
				"Action");
			(void)pnetbuffer_add_string(outdata, fmtstr);
			li = priv_data.filterset_list;
			while (li) {
				fs = slist_get_data(li);
				rc = snprintf(fmtstr,
					256,
					"%-32s %-12s %-12s\n",
					fs->name,
					filterset_type_to_string(fs->type),
					filterset_action_to_string(fs->action));
				(void)pnetbuffer_add_string(outdata, fmtstr);
				li = slist_get_next(li);
			}
			return 0;
		} else {
			(void)pnetbuffer_add_string(outdata,
				"2 Invalid module command\n");
			return 0;
		}

	} else if (!strcmp(argv[i], "stage")) {
		i++;
		if (!strcmp(argv[i], "attach")) {
			if ((argc-i) < 4) {
				(void)pnetbuffer_add_string(outdata,
					"2 Not enough arguments for command\n");
				return 0;
			}
			// name
			i++;
			if (strcmp(argv[i], "name")) {
				(void)pnetbuffer_add_string(outdata,
					"2 Missing name argument\n");
				return 0;
			}
			i++;
			for(state=0 ; state<_MILTER_STATE_COUNT ; state++) {
				if (!strcmp(milter_state_name[state], argv[i]))
					break;
			}
			if (state == _MILTER_STATE_COUNT) {
				(void)pnetbuffer_add_string(outdata,
					"5 An invalid state name was specified\n");
				return 0;
			}
			// filterset
			i++;
			if (strcmp(argv[i], "filterset")) {
				(void)pnetbuffer_add_string(outdata,
					"2 Missing filterset argument\n");
				return 0;
			}
			i++;
			fs = slist_get_filterset_by_name(priv_data.filterset_list,
					argv[i]);
			if (!fs) {
				(void)pnetbuffer_add_string(outdata,
					"5 A filterset by that name does not exists\n");
				return 0;
			}
			rc = milter_is_filterset_attached(fs);
			if (rc) {
				(void)pnetbuffer_add_string(outdata,
					"5 That filterset is already attached\n");
				return 0;
			}
			rc = slist_item_add(&priv_data.state_filterset_list[state],
					fs);
			if (rc) {
				(void)pnetbuffer_add_string(outdata,
					"6 Failed to add the filterset to the stage\n");
				return 0;
			}
			(void)pnetbuffer_add_string(outdata,
				"0 Successfully added filterset to the stage\n");
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_NOTICE,
				"Added filterset %s to stage %s",
				fs->name,
				milter_state_name[state]);

		} else if (!strcmp(argv[i], "detach")) {
			if ((argc-i) < 4) {
				(void)pnetbuffer_add_string(outdata,
					"2 Not enough arguments for command\n");
				return 0;
			}
			// name
			i++;
			if (strcmp(argv[i], "name")) {
				(void)pnetbuffer_add_string(outdata,
					"2 Missing name argument\n");
				return 0;
			}
			i++;
			for(state=0 ; state<_MILTER_STATE_COUNT ; state++) {
				if (!strcmp(milter_state_name[state], argv[i]))
					break;
			}
			if (state == _MILTER_STATE_COUNT) {
				(void)pnetbuffer_add_string(outdata,
					"5 An invalid state name was specified\n");
				return 0;
			}
			// filterset
			i++;
			if (strcmp(argv[i], "filterset")) {
				(void)pnetbuffer_add_string(outdata,
					"2 Missing filterset argument\n");
				return 0;
			}
			i++;
			fs = slist_get_filterset_by_name(priv_data.filterset_list,
					argv[i]);
			if (!fs) {
				(void)pnetbuffer_add_string(outdata,
					"5 A filterset by that name does not exists\n");
				return 0;
			}
			rc = milter_is_filterset_attached(fs);
			if (!rc) {
				(void)pnetbuffer_add_string(outdata,
					"5 That filterset is not attached\n");
				return 0;
			}
			rc = slist_item_remove(&priv_data.state_filterset_list[state],
					fs);
			if (rc) {
				(void)pnetbuffer_add_string(outdata,
					"6 Failed to remove the filterset from"
					" the stage\n");
				return 0;
			}
			(void)pnetbuffer_add_string(outdata,
				"0 Successfully removed filterset from the stage\n");
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_NOTICE,
				"Removed filterset %s from stage %s",
				fs->name,
				milter_state_name[state]);

		} else if (!strcmp(argv[i], "list")) {
			for (i=0 ; i<_MILTER_STATE_COUNT ; i++) {
				rc = snprintf(fmtstr,
					256,
					"%s\n",
					milter_state_name[i]);
				(void)pnetbuffer_add_string(outdata,
					fmtstr);
			}

		} else if (!strcmp(argv[i], "show")) {
			if ((argc-i) < 2) {
				(void)pnetbuffer_add_string(outdata,
					"2 Not enough arguments for command\n");
				return 0;
			}
			// name
			i++;
			if (strcmp(argv[i], "name")) {
				(void)pnetbuffer_add_string(outdata,
					"2 Missing name argument\n");
				return 0;
			}
			i++;
			for(state=0 ; state<_MILTER_STATE_COUNT ; state++) {
				if (!strcmp(milter_state_name[state], argv[i]))
					break;
			}
			if (state == _MILTER_STATE_COUNT) {
				(void)pnetbuffer_add_string(outdata,
					"5 An invalid stage name was specified\n");
				return 0;
			}
			li = priv_data.state_filterset_list[state];
			while (li) {
				fs = slist_get_data(li);
				rc = snprintf(fmtstr,
					256,
					"%s\n",
					fs->name);
				(void)pnetbuffer_add_string(outdata, fmtstr);
				li = slist_get_next(li);
			}

		} else {
		(void)pnetbuffer_add_string(outdata,
			"1 Invalid arguments for command\n");
		return 0;
		}

	} else if (!strcmp(argv[i], "test")) {
		(void)pnetbuffer_add_string(outdata, "0 yep it works");
		return 0;

	} else {
		(void)pnetbuffer_add_string(outdata, "2 invalid module command");
		return 0;
	}
	return 0;
}


int pnetbuffer_sendto(pnetbuffer_t *nb,
			int fd,
			int flags,
			struct sockaddr *sa,
			socklen_t salen)
{
	int		rc;
	uint16_t	plen;
	char		*p;

	pnetbuffer_get_readable(nb, &p, &plen);
	rc = sendto(fd, p, plen, flags, sa, salen);
	return rc;
}



int milter_request_read_header(pnetbuffer_t *indata,
			uint32_t *dlen,
			uint8_t *cmd)
{
	int		rc;
	rc = pnetbuffer_read_uint32(indata, dlen);
	if (rc != 0) return rc;
	rc = pnetbuffer_read_uint8(indata, cmd);
	return rc;
}


int milter_close_client_connection(int fd, milter_metric_index reason_metric)
{
	int			rc;
	ioctl_opt_register_fd_t	opt;
	milter_conn_t		*mc;

	// close connection
	opt.fd = fd;
	opt.port = 0;
	opt.callback = NULL;
	opt.callback_arg = NULL;
	rc = pfm_vector_ioctl(priv_data.module,
		PFM_IOCTL_UNREGISTER_BACKEND_FD,
		&opt);
	rc = tcp_shutdown(fd);

	// cleanup connection state
	mc = &priv_data.conn_table[fd];
	mc->fd = -1;
	mc->io_state = CONN_IO_S_READ_PROTO;
	mc->rem_read_bytes = 0;
	memset(&mc->ctx, 0, sizeof(mail_context_t));
	memset(&mc->fsm, 0, sizeof(microfsm_t));
	pnetbuffer_reset(mc->nb_in);
	pnetbuffer_reset(mc->nb_out);

	// update metrics
	metric_decr_u32(&priv_data.metric[METRIC_CONN_CURRENT]);
	metric_incr_u32(&priv_data.metric[reason_metric]);

	(void)rc; // prevent compiler warning
	return 0;
}



microfsm_result_t milter_state_optneg(microfsm_t *me, microfsm_event_t *event)
{
	int			rc;
	uint8_t			cmd;
	uint32_t		len;
	uint32_t		version;
	uint32_t		u32_1;
	uint32_t		u32_2;
	milter_event_t		*e = (milter_event_t *)event;
	milter_conn_t		*mc = me->userdata;
	pnetbuffer_t		*in = e->nb_in;
	pnetbuffer_t		*out = e->nb_out;

	(void)rc; // prevent compiler warning
	switch(event->signal) {
		case MICROFSM_SIG_ENTRY:
			pfm_vector_log(priv_data.module, LOG_LEVEL_DEBUG,
				"%s() entry", __func__);
			return MICROFSM_RET_HANDLED();
			break;
		case MICROFSM_SIG_EXIT:
			pfm_vector_log(priv_data.module, LOG_LEVEL_DEBUG,
				"%s() exit", __func__);
			return MICROFSM_RET_HANDLED();
			break;
		case SIG_MILTER_REQUEST:
			cmd = 0;
			len = 0;
			rc = pnetbuffer_read_uint32(in, &len);
			rc = pnetbuffer_read_uint8(in, &cmd);
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() SIG_MILTER_REQUEST len=%d cmd=%c",
				__func__, len, cmd);

			if (cmd == SMFIC_OPTNEG) {
				rc = pnetbuffer_read_uint32(in, &version);
				rc = pnetbuffer_read_uint32(in, &u32_1);
				rc = pnetbuffer_read_uint32(in, &u32_2);

				// TODO process optneg request

				rc = pnetbuffer_add_uint32(out, 13);
				rc = pnetbuffer_add_uint8(out, SMFIR_OPTNEG);
				rc = pnetbuffer_add_uint32(out, 2);
				rc = pnetbuffer_add_uint32(out,
					(SMFIF_ADDHDRS|SMFIF_CHGHDRS|SMFIF_QUARANTINE));
				rc = pnetbuffer_add_uint32(out,
					0);
//					(SMFIP_NOBODY));
				return MICROFSM_RET_TRANSITION(&milter_state_connect);

			} else if (cmd == SMFIC_ABORT) {
				// TODO state cleanup and no response
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_NOTICE,
					"Client requested to abort during optneg"
					" fd=%d closing connection",
					mc->fd);
				rc = milter_close_client_connection(mc->fd,
					METRIC_CONN_PROTOCOL_VIOLATION);
				return MICROFSM_RET_HANDLED();

			} else if (cmd == SMFIC_QUIT) {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_INFO,
					"Client requested to close milter connection");
				pnetbuffer_reset(out);
				rc = milter_close_client_connection(mc->fd,
					METRIC_CONN_CLIENT_DROP);
				return MICROFSM_RET_HANDLED();

			} else {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_WARNING,
					"Client protocol violation fd=%d"
					" closing connection",
					mc->fd);
				rc = milter_close_client_connection(mc->fd,
					METRIC_CONN_PROTOCOL_VIOLATION);
				return MICROFSM_RET_HANDLED();
			}

			return MICROFSM_RET_HANDLED();
			break;
	}
	return MICROFSM_RET_IGNORED();
}


microfsm_result_t milter_state_connect(microfsm_t *me, microfsm_event_t *event)
{
	int			rc;
	uint8_t			cmd;
	uint8_t			family;
	uint8_t			macro_cmd;
	uint16_t		port;
	uint32_t		len;
	char			name[MAIL_CONTEXT_STRING_SIZE];
	char			value[MAIL_CONTEXT_STRING_SIZE];
	SList			*item;
	filterset_t		*fs;
	milter_event_t		*e = (milter_event_t *)event;
	milter_conn_t		*mc = me->userdata;
	pnetbuffer_t		*in = e->nb_in;
	pnetbuffer_t		*out = e->nb_out;

	macro_cmd = 0; // prevent compiler warning
	switch(event->signal) {
		case MICROFSM_SIG_ENTRY:
			pfm_vector_log(priv_data.module, LOG_LEVEL_DEBUG,
				"%s() entry", __func__);
			memset(&mc->ctx, 0, sizeof(mail_context_t));
			return MICROFSM_RET_HANDLED();
			break;
		case MICROFSM_SIG_EXIT:
			pfm_vector_log(priv_data.module, LOG_LEVEL_DEBUG,
				"%s() exit", __func__);
			return MICROFSM_RET_HANDLED();
			break;
		case SIG_MILTER_REQUEST:
			cmd = 0;
			len = 0;
			port = 0;
			family = 0;
			rc = pnetbuffer_read_uint32(in, &len);
			rc = pnetbuffer_read_uint8(in, &cmd);
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() SIG_MILTER_REQUEST len=%d cmd=%c",
				__func__, len, cmd);

			if (cmd == SMFIC_CONNECT) {
				rc = pnetbuffer_read_string(in,
					mc->ctx.conn_hostname,
					MAIL_CONTEXT_STRING_SIZE);
				rc = pnetbuffer_read_uint8(in, &family);
				rc = pnetbuffer_read_uint16(in, &port);
				rc = pnetbuffer_read_string(in,
					mc->ctx.conn_addr,
					MAIL_CONTEXT_STRING_SIZE);
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_INFO,
					"%s() SIG_MILTER_REQUEST hostname=%s"
					" family=%c port=%d addr=%s",
					__func__,
					mc->ctx.conn_hostname,
					family,
					port,
					mc->ctx.conn_addr);

				// queue response length
				(void)pnetbuffer_add_uint32(out, 1);

				// perform hostname checks
				fs = NULL;
				item = priv_data.state_filterset_list[MILTER_STATE_CONNECT];
				while (item) {
					fs = slist_get_data(item);
					// only test non-V4CIDR sets
					if (fs->type != F_TYPE_V4CIDR) {
						rc = filterset_test(fs,
								mc->ctx.conn_hostname);
						if (rc != F_ACTION_NONE) break;
					}
					item = slist_get_next(item);
				}
				if (rc == F_ACTION_REJECT) {
					rc = pnetbuffer_add_uint8(out, SMFIR_REJECT);
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_NOTICE,
						"Hostname '%s' filter matched during"
						" %s, rejecting message entirely"
						" fs=%s",
						mc->ctx.conn_hostname,
						milter_state_name[MILTER_STATE_CONNECT],
						fs->name);
					return MICROFSM_RET_TRANSITION(&milter_state_hello);
				}

				// perform address checks
				fs = NULL;
				item = priv_data.state_filterset_list[MILTER_STATE_CONNECT];
				while (item) {
					fs = slist_get_data(item);
					// only test V4CIDR sets
					if (fs->type == F_TYPE_V4CIDR &&
					    family == SMFIA_INET) {
						rc = filterset_test(fs,
								mc->ctx.conn_addr);
						if (rc != F_ACTION_NONE) break;
					}
					item = slist_get_next(item);
				}
				if (rc == F_ACTION_REJECT) {
					rc = pnetbuffer_add_uint8(out, SMFIR_REJECT);
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_NOTICE,
						"Address '%s' filter matched during"
						" %s, rejecting message entirely"
						" fs=%s",
						mc->ctx.conn_addr,
						milter_state_name[MILTER_STATE_CONNECT],
						fs->name);
				} else {
					rc = pnetbuffer_add_uint8(out, SMFIR_CONTINUE);
				}

				return MICROFSM_RET_TRANSITION(&milter_state_hello);

			} else if (cmd == SMFIC_MACRO) {
				// TODO loop over macro definitions
				rc = pnetbuffer_read_uint8(in, &macro_cmd);
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_DEBUG,
					"Macro definitions for %c:", macro_cmd);
				while(1) {
					rc = pnetbuffer_read_string(in,
						name,
						MAIL_CONTEXT_STRING_SIZE);
					if (rc) break;
					rc = pnetbuffer_read_string(in,
						value,
						MAIL_CONTEXT_STRING_SIZE);
					if (rc) break;
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_DEBUG,
						"  name='%s' value='%s'",
						name,
						value);
				}
				// no response for macro request
				return MICROFSM_RET_HANDLED();

			} else if (cmd == SMFIC_ABORT) {
				// TODO state cleanup and no response
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_NOTICE,
					"Client requested to abort during connect"
					" fd=%d closing connection",
					mc->fd);
				rc = milter_close_client_connection(mc->fd,
					METRIC_CONN_PROTOCOL_VIOLATION);
				return MICROFSM_RET_HANDLED();

//				// an abort here is normal for conn shutdown
//				pfm_vector_log(priv_data.module,
//					LOG_LEVEL_DEBUG,
//					"Client requested to abort transaction");
//				// return to HELO state
//				return MICROFSM_RET_TRANSITION(&milter_state_hello);

			} else if (cmd == SMFIC_QUIT) {
				// a close here is normal for conn shutdown
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_INFO,
					"Client requested to close milter connection");
				pnetbuffer_reset(out);
				rc = milter_close_client_connection(mc->fd,
					METRIC_CONN_CLIENT_DROP);
				return MICROFSM_RET_HANDLED();

			} else {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_WARNING,
					"Client protocol violation fd=%d"
					" closing connection",
					mc->fd);
				rc = milter_close_client_connection(mc->fd,
					METRIC_CONN_PROTOCOL_VIOLATION);
				return MICROFSM_RET_HANDLED();
			}

			return MICROFSM_RET_HANDLED();
			break;
	}
	return MICROFSM_RET_IGNORED();
}


microfsm_result_t milter_state_hello(microfsm_t *me, microfsm_event_t *event)
{
	int			rc;
	uint8_t			cmd;
	uint8_t			macro_cmd;
	uint32_t		len;
	char			name[MAIL_CONTEXT_STRING_SIZE];
	char			value[MAIL_CONTEXT_STRING_SIZE];
	SList			*item;
	filter_t		*f;
	filterset_t		*fs;
	milter_event_t		*e = (milter_event_t *)event;
	milter_conn_t		*mc = me->userdata;
	pnetbuffer_t		*in = e->nb_in;
	pnetbuffer_t		*out = e->nb_out;

	macro_cmd = 0; // prevent compiler warning
	switch(event->signal) {
		case MICROFSM_SIG_ENTRY:
			pfm_vector_log(priv_data.module, LOG_LEVEL_DEBUG,
				"%s() entry", __func__);
			memset(&mc->ctx.helo, 0, MAIL_CONTEXT_STRING_SIZE);
			memset(&mc->ctx.mailfrom, 0, MAIL_CONTEXT_STRING_SIZE);
			memset(&mc->ctx.rcptto, 0, MAIL_CONTEXT_STRING_SIZE);
			return MICROFSM_RET_HANDLED();
			break;
		case MICROFSM_SIG_EXIT:
			pfm_vector_log(priv_data.module, LOG_LEVEL_DEBUG,
				"%s() exit", __func__);
			return MICROFSM_RET_HANDLED();
			break;
		case SIG_MILTER_REQUEST:
			cmd = 0;
			len = 0;
			rc = pnetbuffer_read_uint32(in, &len);
			rc = pnetbuffer_read_uint8(in, &cmd);
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() SIG_MILTER_REQUEST len=%d cmd=%c",
				__func__, len, cmd);

			if (cmd == SMFIC_HELO) {
				rc = pnetbuffer_read_string(in,
					mc->ctx.helo,
					MAIL_CONTEXT_STRING_SIZE);
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_INFO,
					"%s() SIG_MILTER_REQUEST helo=%s",
					__func__, mc->ctx.helo);
				metric_incr_u32(&priv_data.metric[METRIC_MAIL_TOTAL]);

				// perform helo checks
				fs = NULL;
				f = NULL;
				item = priv_data.state_filterset_list[MILTER_STATE_HELO];
				while (item) {
					fs = slist_get_data(item);
					f = filterset_match(fs, mc->ctx.helo);
					if (f) break;
					item = slist_get_next(item);
				}
				(void)pnetbuffer_add_uint32(out, 1);
				if (f) {
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_DEBUG,
						"%s() match found for"
						" filterset %s filter %s",
						__func__,
						fs->name,
						f->match);
					if (f->action == F_ACTION_REJECT) {
						rc = pnetbuffer_add_uint8(out, SMFIR_REJECT);
						pfm_vector_log(priv_data.module,
							LOG_LEVEL_NOTICE,
							"Hello '%s' filter"
							" matched during %s,"
							" rejecting message entirely"
							" fs=%s",
							mc->ctx.helo,
							milter_state_name[MILTER_STATE_HELO],
							fs->name);
					} else {
						mc->ctx.score += f->score;
						rc = pnetbuffer_add_uint8(out, SMFIR_CONTINUE);
					}
				} else {
					rc = pnetbuffer_add_uint8(out, SMFIR_CONTINUE);
				}
				return MICROFSM_RET_TRANSITION(&milter_state_mailfrom);

			} else if (cmd == SMFIC_MACRO) {
				// TODO loop over macro definitions
				rc = pnetbuffer_read_uint8(in, &macro_cmd);
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_DEBUG,
					"Macro definitions for %c:", macro_cmd);
				while(1) {
					rc = pnetbuffer_read_string(in,
						name,
						MAIL_CONTEXT_STRING_SIZE);
					if (rc) break;
					rc = pnetbuffer_read_string(in,
						value,
						MAIL_CONTEXT_STRING_SIZE);
					if (rc) break;
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_DEBUG,
						"  name='%s' value='%s'",
						name,
						value);
				}
				// no response for macro request
				return MICROFSM_RET_HANDLED();

			} else if (cmd == SMFIC_ABORT) {
				// TODO state cleanup and no response
				// an abort here is normal for conn shutdown
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_DEBUG,
					"Client requested to abort transaction");
				// return to HELO state
				return MICROFSM_RET_TRANSITION(&milter_state_hello);

			} else if (cmd == SMFIC_QUIT) {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_INFO,
					"Client requested to close milter connection");
				pnetbuffer_reset(out);
				rc = milter_close_client_connection(mc->fd,
					METRIC_CONN_CLIENT_DROP);
				return MICROFSM_RET_HANDLED();

			} else {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_WARNING,
					"Client protocol violation fd=%d"
					" closing connection",
					mc->fd);
				rc = milter_close_client_connection(mc->fd,
					METRIC_CONN_PROTOCOL_VIOLATION);
				return MICROFSM_RET_HANDLED();
			}

			return MICROFSM_RET_HANDLED();
			break;
	}
	return MICROFSM_RET_IGNORED();
}


microfsm_result_t milter_state_mailfrom(microfsm_t *me, microfsm_event_t *event)
{
	int			rc;
	uint8_t			cmd;
	uint8_t			macro_cmd;
	uint32_t		len;
	char			name[MAIL_CONTEXT_STRING_SIZE];
	char			value[MAIL_CONTEXT_STRING_SIZE];
	SList			*item;
	filter_t		*f;
	filterset_t		*fs;
	milter_event_t		*e = (milter_event_t *)event;
	milter_conn_t		*mc = me->userdata;
	pnetbuffer_t		*in = e->nb_in;
	pnetbuffer_t		*out = e->nb_out;

	macro_cmd = 0; // prevent compiler warning
	switch(event->signal) {
		case MICROFSM_SIG_ENTRY:
			pfm_vector_log(priv_data.module, LOG_LEVEL_DEBUG,
				"%s() entry", __func__);
			// reset mail context
			memset(mc->ctx.mailfrom, 0, MAIL_CONTEXT_STRING_SIZE);
			memset(mc->ctx.rcptto, 0, MAIL_CONTEXT_STRING_SIZE);
			mc->ctx.score = 0;
			mc->body_total_len = 0;
			return MICROFSM_RET_HANDLED();
			break;
		case MICROFSM_SIG_EXIT:
			pfm_vector_log(priv_data.module, LOG_LEVEL_DEBUG,
				"%s() exit", __func__);
			return MICROFSM_RET_HANDLED();
			break;
		case SIG_MILTER_REQUEST:
			cmd = 0;
			len = 0;
			rc = pnetbuffer_read_uint32(in, &len);
			rc = pnetbuffer_read_uint8(in, &cmd);
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() SIG_MILTER_REQUEST len=%d cmd=%c",
				__func__, len, cmd);

			if (cmd == SMFIC_MAIL) {
				// TODO loop over addrs
				rc = pnetbuffer_read_string(in,
					mc->ctx.mailfrom,
					MAIL_CONTEXT_STRING_SIZE);
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_INFO,
					"%s() mail from '%s'",
					__func__,
					mc->ctx.mailfrom);

				// perform mailfrom checks
				fs = NULL;
				f = NULL;
				item = priv_data.state_filterset_list[MILTER_STATE_MAILFROM];
				while (item) {
					fs = slist_get_data(item);
					f = filterset_match(fs, mc->ctx.mailfrom);
					if (f) break;
					item = slist_get_next(item);
				}
				(void)pnetbuffer_add_uint32(out, 1);
				if (f) {
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_DEBUG,
						"%s() match found for"
						" filterset %s filter %s",
						__func__,
						fs->name,
						f->match);
					if (f->action == F_ACTION_REJECT) {
						rc = pnetbuffer_add_uint8(out, SMFIR_REJECT);
						pfm_vector_log(priv_data.module,
							LOG_LEVEL_NOTICE,
							"Address '%s' filter"
							" matched during %s,"
							" rejecting message entirely"
							" fs=%s",
							mc->ctx.mailfrom,
							milter_state_name[MILTER_STATE_MAILFROM],
							fs->name);
					} else {
						mc->ctx.score += f->score;
						rc = pnetbuffer_add_uint8(out, SMFIR_CONTINUE);
					}
				} else {
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_DEBUG,
						"%s() no match found in"
						" filterset(s), continuing"
						" with transaction",
						__func__);
					rc = pnetbuffer_add_uint8(out, SMFIR_CONTINUE);
				}
				return MICROFSM_RET_TRANSITION(&milter_state_rcptto);

			} else if (cmd == SMFIC_MACRO) {
				// TODO loop over macro definitions
				rc = pnetbuffer_read_uint8(in, &macro_cmd);
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_DEBUG,
					"Macro definitions for %c:", macro_cmd);
				while(1) {
					rc = pnetbuffer_read_string(in,
						name,
						MAIL_CONTEXT_STRING_SIZE);
					if (rc) break;
					rc = pnetbuffer_read_string(in,
						value,
						MAIL_CONTEXT_STRING_SIZE);
					if (rc) break;
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_DEBUG,
						"  name='%s' value='%s'",
						name,
						value);
				}
				// no response for macro request
				return MICROFSM_RET_HANDLED();

			} else if (cmd == SMFIC_ABORT) {
				// TODO state cleanup and no response
				// an abort here is normal for conn shutdown
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_DEBUG,
					"Client requested to abort transaction");
				// return to HELO state
				return MICROFSM_RET_TRANSITION(&milter_state_hello);

			} else if (cmd == SMFIC_QUIT) {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_INFO,
					"Client requested to close milter connection");
				pnetbuffer_reset(out);
				rc = milter_close_client_connection(mc->fd,
					METRIC_CONN_CLIENT_DROP);
				return MICROFSM_RET_HANDLED();

			} else {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_WARNING,
					"Client protocol violation fd=%d"
					" closing connection",
					mc->fd);
				rc = milter_close_client_connection(mc->fd,
					METRIC_CONN_PROTOCOL_VIOLATION);
				return MICROFSM_RET_HANDLED();
			}

			return MICROFSM_RET_HANDLED();
			break;
	}
	return MICROFSM_RET_IGNORED();
}


microfsm_result_t milter_state_rcptto(microfsm_t *me, microfsm_event_t *event)
{
	int			rc;
	uint8_t			cmd;
	uint8_t			macro_cmd;
	uint32_t		len;
	char			name[MAIL_CONTEXT_STRING_SIZE];
	char			value[MAIL_CONTEXT_STRING_SIZE];
	SList			*item;
	filterset_t		*fs;
	filter_t		*f;
	milter_event_t		*e = (milter_event_t *)event;
	milter_conn_t		*mc = me->userdata;
	pnetbuffer_t		*in = e->nb_in;
	pnetbuffer_t		*out = e->nb_out;

	macro_cmd = 0; // prevent compiler warning
	switch(event->signal) {
		case MICROFSM_SIG_ENTRY:
			pfm_vector_log(priv_data.module, LOG_LEVEL_DEBUG,
				"%s() entry", __func__);
			memset(mc->ctx.rcptto, 0, MAIL_CONTEXT_STRING_SIZE);
			return MICROFSM_RET_HANDLED();
			break;
		case MICROFSM_SIG_EXIT:
			pfm_vector_log(priv_data.module, LOG_LEVEL_DEBUG,
				"%s() exit", __func__);
			return MICROFSM_RET_HANDLED();
			break;
		case SIG_MILTER_REQUEST:
			cmd = 0;
			len = 0;
			rc = pnetbuffer_read_uint32(in, &len);
			rc = pnetbuffer_read_uint8(in, &cmd);
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() SIG_MILTER_REQUEST len=%d cmd=%c",
				__func__, len, cmd);

			if (cmd == SMFIC_RCPT) {
				// TODO loop over addrs
				rc = pnetbuffer_read_string(in,
					mc->ctx.rcptto,
					MAIL_CONTEXT_STRING_SIZE);
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_INFO,
					"%s() rcpt to '%s'",
					__func__,
					mc->ctx.rcptto);

				// perform rcptto checks
				fs = NULL;
				f = NULL;
				item = priv_data.state_filterset_list[MILTER_STATE_RCPTTO];
				while (item) {
					fs = slist_get_data(item);
					f = filterset_match(fs, mc->ctx.rcptto);
					if (f) break;
					item = slist_get_next(item);
				}
				(void)pnetbuffer_add_uint32(out, 1);
				if (f) {
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_DEBUG,
						"%s() match found for"
						" filterset %s filter %s",
						__func__,
						fs->name,
						f->match);
					if (f->action == F_ACTION_REJECT) {
						rc = pnetbuffer_add_uint8(out, SMFIR_REJECT);
						pfm_vector_log(priv_data.module,
							LOG_LEVEL_NOTICE,
							"Address '%s' filter"
							" matched during %s,"
							" rejecting message entirely"
							" fs=%s",
							mc->ctx.rcptto,
							milter_state_name[MILTER_STATE_RCPTTO],
							fs->name);
					} else {
						mc->ctx.score += f->score;
						rc = pnetbuffer_add_uint8(out, SMFIR_CONTINUE);
					}
				} else {
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_DEBUG,
						"%s() no match found in"
						" filterset(s), continuing"
						" with transaction",
						__func__);
					rc = pnetbuffer_add_uint8(out, SMFIR_CONTINUE);
				}
				return MICROFSM_RET_TRANSITION(&milter_state_headers);

			} else if (cmd == SMFIC_MACRO) {
				// TODO loop over macro definitions
				rc = pnetbuffer_read_uint8(in, &macro_cmd);
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_DEBUG,
					"Macro definitions for %c:", macro_cmd);
				while(1) {
					rc = pnetbuffer_read_string(in,
						name,
						MAIL_CONTEXT_STRING_SIZE);
					if (rc) break;
					rc = pnetbuffer_read_string(in,
						value,
						MAIL_CONTEXT_STRING_SIZE);
					if (rc) break;
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_DEBUG,
						"  name='%s' value='%s'",
						name,
						value);
				}
				// no response for macro request
				return MICROFSM_RET_HANDLED();

			} else if (cmd == SMFIC_ABORT) {
				// TODO state cleanup and no response
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_NOTICE,
					"Client requested to abort transaction");
				// return to HELO state
				return MICROFSM_RET_TRANSITION(&milter_state_hello);

			} else if (cmd == SMFIC_QUIT) {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_INFO,
					"Client requested to close milter connection");
				pnetbuffer_reset(out);
				rc = milter_close_client_connection(mc->fd,
					METRIC_CONN_CLIENT_DROP);
				return MICROFSM_RET_HANDLED();

			} else {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_WARNING,
					"Client protocol violation fd=%d"
					" closing connection",
					mc->fd);
				rc = milter_close_client_connection(mc->fd,
					METRIC_CONN_PROTOCOL_VIOLATION);
				return MICROFSM_RET_HANDLED();
			}

			return MICROFSM_RET_HANDLED();
			break;
	}
	return MICROFSM_RET_IGNORED();
}


microfsm_result_t milter_state_headers(microfsm_t *me, microfsm_event_t *event)
{
	int			rc;
	uint8_t			cmd;
	uint8_t			macro_cmd;
	uint32_t		len;
	char			name[MAIL_CONTEXT_STRING_SIZE];
	char			value[MAIL_CONTEXT_STRING_SIZE];
	SList			*item;
	filterset_t		*fs;
	filter_t		*f;
	milter_event_t		*e = (milter_event_t *)event;
	milter_conn_t		*mc = me->userdata;
	pnetbuffer_t		*in = e->nb_in;
	pnetbuffer_t		*out = e->nb_out;

	macro_cmd = 0; // prevent compiler warning
	switch(event->signal) {
		case MICROFSM_SIG_ENTRY:
			pfm_vector_log(priv_data.module, LOG_LEVEL_DEBUG,
				"%s() entry", __func__);
			return MICROFSM_RET_HANDLED();
			break;
		case MICROFSM_SIG_EXIT:
			pfm_vector_log(priv_data.module, LOG_LEVEL_DEBUG,
				"%s() exit", __func__);
			return MICROFSM_RET_HANDLED();
			break;
		case SIG_MILTER_REQUEST:
			cmd = 0;
			len = 0;
			rc = pnetbuffer_read_uint32(in, &len);
			rc = pnetbuffer_read_uint8(in, &cmd);
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() SIG_MILTER_REQUEST len=%d cmd=%c",
				__func__, len, cmd);

			if (cmd == SMFIC_HEADER) {
				rc = pnetbuffer_read_string(in,
					name,
					MAIL_CONTEXT_STRING_SIZE);
				rc = pnetbuffer_read_string(in,
					value,
					MAIL_CONTEXT_STRING_SIZE);
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_DEBUG,
					"header name='%s' value='%s'",
					name,
					value);

				// perform header checks
				fs = NULL;
				f = NULL;
				item = priv_data.state_filterset_list[MILTER_STATE_HEADER];
				while (item) {
					fs = slist_get_data(item);
					f = filterset_match(fs, value);
					if (f) break;
					item = slist_get_next(item);
				}
				(void)pnetbuffer_add_uint32(out, 1);
				if (f) {
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_DEBUG,
						"%s() match found for"
						" filterset %s filter %s",
						__func__,
						fs->name,
						f->match);
					if (f->action == F_ACTION_REJECT) {
						rc = pnetbuffer_add_uint8(out, SMFIR_REJECT);
						pfm_vector_log(priv_data.module,
							LOG_LEVEL_NOTICE,
							"Header '%s' value"
							" '%s' filter matched during"
							" %s, rejecting message"
							" entirely fs=%s",
							name,
							value,
							milter_state_name[MILTER_STATE_HEADER],
							fs->name);
					} else {
						mc->ctx.score += f->score;
						rc = pnetbuffer_add_uint8(out, SMFIR_CONTINUE);
					}
				} else {
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_DEBUG,
						"%s() no match found in"
						" filterset(s), continuing"
						" with transaction",
						__func__);
					rc = pnetbuffer_add_uint8(out, SMFIR_CONTINUE);
				}
				return MICROFSM_RET_HANDLED();

			} else if (cmd == SMFIC_EOH) {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_INFO,
					"%s() SIG_MILTER_REQUEST header"
					" processing complete",
					__func__);
				rc = pnetbuffer_add_uint32(out, 1);
				rc = pnetbuffer_add_uint8(out, SMFIR_CONTINUE);
				return MICROFSM_RET_TRANSITION(&milter_state_body);

			} else if (cmd == SMFIC_MACRO) {
				// TODO loop over macro definitions
				rc = pnetbuffer_read_uint8(in, &macro_cmd);
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_DEBUG,
					"Macro definitions for %c:", macro_cmd);
				while(1) {
					rc = pnetbuffer_read_string(in,
						name,
						MAIL_CONTEXT_STRING_SIZE);
					if (rc) break;
					rc = pnetbuffer_read_string(in,
						value,
						MAIL_CONTEXT_STRING_SIZE);
					if (rc) break;
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_DEBUG,
						"  name='%s' value='%s'",
						name,
						value);
				}
				// no response for macro request
				return MICROFSM_RET_HANDLED();

			} else if (cmd == SMFIC_ABORT) {
				// TODO state cleanup and no response
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_NOTICE,
					"Client requested to abort transaction");
				// return to HELO state
				return MICROFSM_RET_TRANSITION(&milter_state_hello);

			} else if (cmd == SMFIC_QUIT) {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_INFO,
					"Client requested to close milter connection");
				pnetbuffer_reset(out);
				rc = milter_close_client_connection(mc->fd,
					METRIC_CONN_CLIENT_DROP);
				return MICROFSM_RET_HANDLED();

			} else {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_WARNING,
					"Client protocol violation fd=%d"
					" closing connection",
					mc->fd);
				rc = milter_close_client_connection(mc->fd,
					METRIC_CONN_PROTOCOL_VIOLATION);
				return MICROFSM_RET_HANDLED();
			}

			return MICROFSM_RET_HANDLED();
			break;
	}
	return MICROFSM_RET_IGNORED();
}


microfsm_result_t milter_state_body(microfsm_t *me, microfsm_event_t *event)
{
	int			rc;
	filter_action		vote;
	uint8_t			cmd;
	uint8_t			macro_cmd;
	uint8_t			result;
	uint32_t		len;
	uint16_t		blen;
	char			name[MAIL_CONTEXT_STRING_SIZE];
	char			value[MAIL_CONTEXT_STRING_SIZE];
	char			*body;
	milter_event_t		*e = (milter_event_t *)event;
	milter_conn_t		*mc = me->userdata;
	pnetbuffer_t		*in = e->nb_in;
	pnetbuffer_t		*out = e->nb_out;

	macro_cmd = 0; // prevent compiler warning
	switch(event->signal) {
		case MICROFSM_SIG_ENTRY:
			pfm_vector_log(priv_data.module, LOG_LEVEL_DEBUG,
				"%s() entry", __func__);
			return MICROFSM_RET_HANDLED();
			break;
		case MICROFSM_SIG_EXIT:
			pfm_vector_log(priv_data.module, LOG_LEVEL_DEBUG,
				"%s() exit", __func__);
			return MICROFSM_RET_HANDLED();
			break;

		case SIG_MILTER_RAW_DATA:
			pnetbuffer_get_readable(in, &body, &blen);

			mc->body_total_len += blen;

			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() SIG_MILTER_RAW_DATA bodylen=%d",
				__func__, blen);

			// DEBUG line
//			dump_memory_hex(body, blen);

			// TODO process body chunk

			if (mc->rem_read_bytes == 0) {
				// send the packet response
				// if there are no more body chunks
				rc = pnetbuffer_add_uint32(out, 1);
				rc = pnetbuffer_add_uint8(out, SMFIR_CONTINUE);
			}

			return MICROFSM_RET_HANDLED();
			break;

		case SIG_MILTER_REQUEST:
			cmd = 0;
			len = 0;

			// DEBUG lines
			pnetbuffer_get_readable(in, &body, &blen);
//			dump_memory_hex(body, blen);

			rc = pnetbuffer_read_uint32(in, &len);
			rc = pnetbuffer_read_uint8(in, &cmd);
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() SIG_MILTER_REQUEST len=%d cmd=%c",
				__func__, len, cmd);

			if (cmd == SMFIC_BODY) {
				pnetbuffer_get_readable(in, &body, &blen);

				mc->body_total_len += blen;

				// TODO filterset_test needs to support binary data
				if (blen > 0)
					body[blen-1] = 0;
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_DEBUG,
					"%s() SIG_MILTER_REQUEST bodylen=%d",
					__func__, blen);

				// perform body checks
/*				fs = NULL;
				item = priv_data.state_filterset_list[MILTER_STATE_BODY];
				while (item) {
					fs = slist_get_data(item);
					// TODO filterset_test needs to
					// support binary data
					rc = filterset_test(fs, body);
					if (rc != F_ACTION_NONE) break;
					item = slist_get_next(item);
				}
				(void)pnetbuffer_add_uint32(out, 1);
				if (rc == F_ACTION_REJECT) {
					rc = pnetbuffer_add_uint8(out, SMFIR_REJECT);
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_NOTICE,
						"Filter matched during"
						" %s, rejecting message entirely"
						" fs=%s",
						milter_state_name[MILTER_STATE_BODY],
						fs->name);
				} else {
					rc = pnetbuffer_add_uint8(out, SMFIR_CONTINUE);
				}
*/

				if (mc->rem_read_bytes == 0) {
					// send the packet response
					// if there are no more body chunks
					rc = pnetbuffer_add_uint32(out, 1);
					rc = pnetbuffer_add_uint8(out, SMFIR_CONTINUE);
				}

				return MICROFSM_RET_HANDLED();

			} else if (cmd == SMFIC_BODYEOB) {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_INFO,
					"%s() SIG_MILTER_REQUEST body"
					" processing complete",
					__func__);
				// TODO final checks
				result = SMFIR_CONTINUE;
				vote = F_ACTION_ACCEPT;
				if (mc->ctx.score >= priv_data.reject_score)
					vote = F_ACTION_REJECT;
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_INFO,
					"Mail final state host=%s addr=%s"
					" helo=%s from=%s rcpt=%s bodylen=%d"
					" score=%d action=%s",
					mc->ctx.conn_hostname,
					mc->ctx.conn_addr,
					mc->ctx.helo,
					mc->ctx.mailfrom,
					mc->ctx.rcptto,
					mc->body_total_len,
					mc->ctx.score,
					filterset_action_to_string(vote));

				if (vote == F_ACTION_REJECT) {
					result = SMFIR_REJECT;
					metric_incr_u32(&priv_data.metric[METRIC_MAIL_REJECT]);
				} else {
					result = SMFIR_CONTINUE;
					metric_incr_u32(&priv_data.metric[METRIC_MAIL_PASS]);
				}
				rc = pnetbuffer_add_uint32(out, 1);
				rc = pnetbuffer_add_uint8(out, result);
				// return to MAIL state
				return MICROFSM_RET_TRANSITION(&milter_state_mailfrom);

			} else if (cmd == SMFIC_MACRO) {
				// TODO loop over macro definitions
				rc = pnetbuffer_read_uint8(in, &macro_cmd);
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_DEBUG,
					"Macro definitions for %c:", macro_cmd);
				while(1) {
					rc = pnetbuffer_read_string(in,
						name,
						MAIL_CONTEXT_STRING_SIZE);
					if (rc) break;
					rc = pnetbuffer_read_string(in,
						value,
						MAIL_CONTEXT_STRING_SIZE);
					if (rc) break;
					pfm_vector_log(priv_data.module,
						LOG_LEVEL_DEBUG,
						"  name='%s' value='%s'",
						name,
						value);
				}
				// no response for macro request
				return MICROFSM_RET_HANDLED();

			} else if (cmd == SMFIC_ABORT) {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_NOTICE,
					"Client requested to abort transaction");
				// TODO state cleanup and no response
				// return to HELO state
				return MICROFSM_RET_TRANSITION(&milter_state_hello);

			} else if (cmd == SMFIC_QUIT) {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_INFO,
					"Client requested to close milter connection");
				pnetbuffer_reset(out);
				rc = milter_close_client_connection(mc->fd,
					METRIC_CONN_CLIENT_DROP);
				return MICROFSM_RET_HANDLED();

			} else {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_WARNING,
					"Client protocol violation fd=%d"
					" closing connection",
					mc->fd);
				rc = milter_close_client_connection(mc->fd,
					METRIC_CONN_PROTOCOL_VIOLATION);
				return MICROFSM_RET_HANDLED();
			}

			return MICROFSM_RET_HANDLED();
			break;
	}
	return MICROFSM_RET_IGNORED();
}


int socket_enable_write_notify(int fd)
{
	int			rc;
	ioctl_opt_modify_fd_t	modopt;

	// notify when socket becomes writable
	modopt.fd = fd;
	modopt.flags = PFM_FD_MODIFY_F_WRITABLE;
	rc = pfm_vector_ioctl(priv_data.module,
			PFM_IOCTL_MODIFY_BACKEND_FD,
			&modopt);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_WARNING,
		"pfm_vector_ioctl() rc=%d"
		" failed to modify fd (notify writable)",
		rc);
	return rc;
}


int socket_disable_write_notify(int fd)
{
	int			rc;
	ioctl_opt_modify_fd_t	modopt;

	// do not notify when socket becomes writable
	modopt.fd = fd;
	modopt.flags = 0;
	rc = pfm_vector_ioctl(priv_data.module,
			PFM_IOCTL_MODIFY_BACKEND_FD,
			&modopt);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_WARNING,
		"pfm_vector_ioctl() rc=%d"
		" failed to modify fd (dont notify writable)",
		rc);
	return rc;
}


int send_queued_response_data(int fd, milter_conn_t *mc)
{
	int			rc;
	uint16_t		plen;
	char			*p;

	// send response packet
	pnetbuffer_get_readable(mc->nb_out, &p, &plen);
//	dump_memory_hex(p, plen);
	if (plen) {
		rc = send(fd, p, plen, 0);
		if (rc == -1) {
			rc = errno;
			if (rc == EAGAIN || rc == EWOULDBLOCK) {
				// wait for socket to become writable
				rc = socket_enable_write_notify(fd);
				return rc;
			}
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_WARNING,
				"send() rc=%d failed to write"
				" response packet,"
				" closing connection",
				rc);
			(void)milter_close_client_connection(fd,
				METRIC_CONN_SERVER_DROP);
			return rc;
		}
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"send() len=%d/%d",
			rc,
			plen);
		plen = rc;
		rc = pnetbuffer_read_advance(mc->nb_out, plen);
		if (rc) {
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_WARNING,
				"pnetbuffer_read_advance() rc=%d"
				" failed to advance read position,"
				" closing connection",
				rc);
			(void)milter_close_client_connection(fd,
				METRIC_CONN_SERVER_DROP);
			return rc;
		}
	}
	pnetbuffer_reset(mc->nb_out);
	return 0;
}


int process_request_packets(int fd, milter_conn_t *mc)
{
	int			rc;
	uint16_t		plen;
	uint16_t		blen;
	uint32_t		mp_len;
	char			*p;
	pnetbuffer_t		*nb_in;
	milter_event_t		event;

	// TODO find a better way to split protocol packets up
	// than to copy to a scratch buffer
	pnetbuffer_t		*nb_temp;
	char			buffer_temp[PACKET_BUFFER_SIZE];

	nb_in = mc->nb_in;

	// handle multiple packets in the buffer (MACRO/etc)
	nb_temp = pnetbuffer_init(buffer_temp, PACKET_BUFFER_SIZE);
	pnetbuffer_get_readable(mc->nb_in, &p, &blen);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() processing read buffer len=%d",
		__func__,
		blen);
	while (blen > 0) {
		mp_len = 0;
		rc = pnetbuffer_peek_uint32(nb_in, &mp_len);
		if (rc == PNETBUFFER_RC_NOTENOUGH) {
			// wait for more packet data to arrive
			// TODO maybe collapse nb_in here?
			return 0;
		}
		if (mp_len > blen) {
			// switch to chunk mode
			mc->io_state = CONN_IO_S_READ_RAW;
			mc->rem_read_bytes = mp_len - (blen-4);
			mp_len = blen-4; // 1st body packet w/ header
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"%s() chunk mode rembytes=%d mp_len=%d",
				__func__,
				mc->rem_read_bytes,
				mp_len);
		}
		pnetbuffer_reset(nb_temp);
		pnetbuffer_get_writable(nb_temp, &p, &plen);
		rc = pnetbuffer_read_data(nb_in, p, (uint16_t)mp_len+4);
		rc = pnetbuffer_set_written(nb_temp, (uint16_t)mp_len+4);
		blen -= (mp_len+4);

		// FSM event/dispatch
		event.super.signal = SIG_MILTER_REQUEST;
		event.nb_in = nb_temp;
		event.nb_out = mc->nb_out;
		rc = microfsm_dispatch(&mc->fsm, (microfsm_event_t *)&event);
		if (mc->fd == -1) {
			// the connection was closed (proto violation, etc)
			// cant send a response
			return 0;
		}

		// send response packet
		rc = send_queued_response_data(fd, mc);
		if (rc)
			return rc;
	}

	// reset buffer if possible
	pnetbuffer_get_readable(mc->nb_in, &p, &blen);
	if (blen == 0)
		pnetbuffer_reset(mc->nb_in);

	return 0;
}


int process_chunk_data(int fd, milter_conn_t *mc)
{
	int			rc;
	uint16_t		plen;
	uint16_t		blen;
	uint32_t		mp_len;
	char			*p;
	pnetbuffer_t		*nb_in;
	milter_event_t		event;

	// TODO find a better way to split protocol packets up
	// than to copy to a scratch buffer
	pnetbuffer_t		*nb_temp;
	char			buffer_temp[PACKET_BUFFER_SIZE];

	nb_in = mc->nb_in;

	// handle multiple packets in the buffer (MACRO/etc)
	nb_temp = pnetbuffer_init(buffer_temp, PACKET_BUFFER_SIZE);
	blen = 0;
	pnetbuffer_get_readable(mc->nb_in, &p, &blen);
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"%s() processing read buffer len=%d",
		__func__,
		blen);

	if (blen < mc->rem_read_bytes) {
		// buffer has another chunk, more to come
		mp_len = blen;

	} else if (blen > mc->rem_read_bytes) {
		// buffer has last chunk and protocol packet(s)
		mc->io_state = CONN_IO_S_READ_PROTO;
		mp_len = mc->rem_read_bytes;

	} else {
		// buffer has just the last chunk
		mc->io_state = CONN_IO_S_READ_PROTO;
		mp_len = blen;
	}
	mc->rem_read_bytes -= mp_len;

	pnetbuffer_reset(nb_temp);
	pnetbuffer_get_writable(nb_temp, &p, &plen);
	rc = pnetbuffer_read_data(nb_in, p, (uint16_t)mp_len);
	rc = pnetbuffer_set_written(nb_temp, (uint16_t)mp_len);

	// FSM event/dispatch
	event.super.signal = SIG_MILTER_RAW_DATA;
	event.nb_in = nb_temp;
	event.nb_out = mc->nb_out;
	rc = microfsm_dispatch(&mc->fsm, (microfsm_event_t *)&event);

	// reset buffer if possible
	pnetbuffer_get_readable(mc->nb_in, &p, &blen);
	if (blen == 0)
		pnetbuffer_reset(mc->nb_in);

	// send response packet, if the socket wasn't closed
	// inside the state machine
	if (mc->fd != -1)
		rc = send_queued_response_data(fd, mc);

	return rc;
}


int milter_io_callback(int fd, int events, connection_data_t *cd, void *arg)
{
	int			rc;
	uint16_t		plen;
	int			read_len;
	socklen_t		saddr_len;
	char			*p;

	pnetbuffer_t		*nb_in;
	pnetbuffer_t		*nb_out;

	io_context_t		ctx;
	milter_conn_t		*mc;

	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"milter_io_callback() entry");

	//************************************
	// init
	//************************************
	mc = arg;
	ctx.dest_port = cd->port;
	nb_in = mc->nb_in;
	nb_out = mc->nb_out;

	//************************************
	// handle socket exception
	//************************************
	if (events & EPOLLHUP) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"HUP socket exception occurred"
			" closing client connection");
		rc = milter_close_client_connection(fd,
			METRIC_CONN_SERVER_DROP);
		return 0;
	}

	//************************************
	// read packet
	//************************************
	if (events & EPOLLIN) {
		pnetbuffer_get_writable(nb_in, &p, &plen);
		saddr_len = sizeof(struct sockaddr);
		read_len = recvfrom(fd,
				p,
				(int)plen,
				0,
				&ctx.src_addr,
				&saddr_len);
		if (read_len == -1) {
			// an error occurred during a socket read attempt
			rc = errno;
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_WARNING,
				"recv() rc=%d closing client connection",
				rc);
			(void)milter_close_client_connection(fd,
				METRIC_CONN_SERVER_DROP);
			return rc;
		}
		if (read_len == 0) {
			// the client closed their end of the connection
			rc = errno;
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_INFO,
				"recv() rc=%d client disconnected closing connection",
				rc);
			(void)milter_close_client_connection(fd,
				METRIC_CONN_CLIENT_DROP);
			return rc;
		}
		rc = pnetbuffer_set_written(nb_in, (uint16_t)read_len);
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_DEBUG,
			"recv() read_len=%d",
			read_len);
//		dump_memory_hex(p, read_len);

		if (mc->io_state == CONN_IO_S_READ_PROTO) {
			// handle protocol packets
			rc = process_request_packets(fd, mc);
			return rc;

		} else if (mc->io_state == CONN_IO_S_READ_RAW) {
			// handle raw data (body continuation)
			rc = process_chunk_data(fd, mc);

			// if there is remaining data in nb_in,
			// it is packet data
			pnetbuffer_get_readable(mc->nb_in, &p, &plen);
			if (plen == 0) {
				rc = process_request_packets(fd, mc);
				return rc;
			}

		} else {
			// bad state

		}
	}

	//************************************
	// write queued packet data
	//************************************
	if (events & EPOLLOUT) {
		pnetbuffer_get_readable(nb_out, &p, &plen);
//		dump_memory_hex(p, plen);
		if (plen) {
			// send more data
			rc = send(fd, p, plen, 0);
			if (rc == -1) {
				rc = errno;
				if (rc == EAGAIN || rc == EWOULDBLOCK) {
					// wait for socket to become writable
					rc = socket_enable_write_notify(fd);
					return rc;
				}
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_WARNING,
					"send() rc=%d failed to write"
					" response data,"
					" closing connection",
					rc);
				(void)milter_close_client_connection(fd,
					METRIC_CONN_SERVER_DROP);
				return rc;
			}
			pfm_vector_log(priv_data.module,
				LOG_LEVEL_DEBUG,
				"send() len=%d/%d",
				rc,
				plen);
			plen = rc;
			rc = pnetbuffer_read_advance(nb_out, plen);
			if (rc) {
				pfm_vector_log(priv_data.module,
					LOG_LEVEL_WARNING,
					"pnetbuffer_read_advance() rc=%d"
					" failed to advance read position,"
					" closing connection",
					rc);
				(void)milter_close_client_connection(fd,
					METRIC_CONN_SERVER_DROP);
				return rc;
			}
		} else {
			// no more data to send, stop fd writable events
			rc = socket_disable_write_notify(fd);
			return rc;
		}
	}

	return 0;
}


int server_io_callback(int fd, int events, connection_data_t *cd, void *arg)
{
	int			rc;
	int			c_flags;
	int			c_fd;
	socklen_t		c_len;
	struct sockaddr		c_sa;
	ioctl_opt_register_fd_t	opt;
	milter_conn_t		*mc;

	metric_incr_u32(&priv_data.metric[METRIC_CONN_TOTAL]);
	c_len = sizeof(struct sockaddr);
	c_fd = accept(fd, &c_sa, &c_len);
	if (c_fd == -1) {
		rc = errno;
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"Failed to accept new client connection rc=%d",
			rc);
		metric_incr_u32(&priv_data.metric[METRIC_CONN_SERVER_DROP]);
		return rc;
	}
	c_flags = fcntl(c_fd, F_GETFL, 0);
	rc = fcntl(c_fd, F_SETFL, c_flags | O_NONBLOCK);
	if (rc == -1) {
		rc = errno;
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"Failed to set O_NONBLOCK on new client"
			" connection rc=%d",
			rc);
		c_flags = tcp_shutdown(c_fd);
		metric_incr_u32(&priv_data.metric[METRIC_CONN_SERVER_DROP]);
		return rc;
	}

	// setup milter connection
	if (c_fd >= CONN_TABLE_SIZE_DEFAULT) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"new client connection denied (too many)");
		c_flags = tcp_shutdown(c_fd);
		metric_incr_u32(&priv_data.metric[METRIC_CONN_SERVER_DROP]);
		return rc;
	}
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_DEBUG,
		"Setting up new client connection");
	mc = &priv_data.conn_table[c_fd];
	mc->fd = c_fd;
	mc->io_state = CONN_IO_S_READ_PROTO;
	mc->rem_read_bytes = 0;
	memset(&mc->ctx, 0, sizeof(mail_context_t));
	pnetbuffer_reset(mc->nb_in);
	pnetbuffer_reset(mc->nb_out);
	rc = microfsm_init(&mc->fsm, milter_state_optneg);
	if (rc) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_WARNING,
			"new client connection denied (connection setup error)");
		c_flags = tcp_shutdown(c_fd);
		metric_incr_u32(&priv_data.metric[METRIC_CONN_SERVER_DROP]);
		return rc;
	}
	rc = microfsm_set_userdata(&mc->fsm, mc);

	opt.fd = c_fd;
	opt.port = 0;
	opt.callback = milter_io_callback;
	opt.callback_arg = mc;
	rc = pfm_vector_ioctl(priv_data.module,
		PFM_IOCTL_REGISTER_BACKEND_FD,
		&opt);
	if (rc) {
		pfm_vector_log(priv_data.module,
			LOG_LEVEL_ERROR,
			"Failed to register new client FD with ioloop"
			" subsystem rc=%d",
			rc);
		c_flags = tcp_shutdown(c_fd);
		metric_incr_u32(&priv_data.metric[METRIC_CONN_SERVER_DROP]);
		return rc;
	}
	pfm_vector_log(priv_data.module,
		LOG_LEVEL_INFO,
		"fd=%d Accepted new client connection",
		c_fd);
	metric_incr_u32(&priv_data.metric[METRIC_CONN_CURRENT]);

	return 0;
}


int vector_handle_new_connection(pfm_t *pfm,
			io_context_t *ctx,
			pnetbuffer_t *indata,
			pnetbuffer_t *outdata)
{
	// nothing to do here now
	return 0;
}


int dump_config_filterset_add_entry(filter_t *f, void *arg)
{
	int			rc;
	char			buffer[LINEBUFFER_SIZE];
	filterset_enum_arg_t	*ea = arg;

	rc = snprintf (buffer,
		LINEBUFFER_SIZE,
		"pfm command milter filterset entry add name %s"
		" match %s score %d\n",
		ea->fs->name,
		f->match,
		f->score);
	rc = pnetbuffer_add_string(ea->nb, buffer);
	if (rc) return rc;
	return 0;
}


int milter_dump_config(pnetbuffer_t *nb_out)
{
	int			rc;
	int			i;
	SList			*li;
	filterset_t		*fs;
	char			buffer[LINEBUFFER_SIZE];
	filterset_enum_arg_t	ea;

	rc = snprintf (buffer,
			LINEBUFFER_SIZE,
			"pfm command milter policy set reject_score %d\n",
			priv_data.reject_score);
	rc = pnetbuffer_add_string(nb_out, buffer);
	if (rc) return rc;


	li = priv_data.filterset_list;
	while (li) {
		fs = slist_get_data(li);
		rc = snprintf (buffer,
				LINEBUFFER_SIZE,
				"pfm command milter filterset create name %s"
				" type %s action %s\n",
				fs->name,
				filterset_type_to_string(fs->type),
				filterset_action_to_string(fs->action));
		rc = pnetbuffer_add_string(nb_out, buffer);
		if (rc) return rc;

		ea.fs = fs;
		ea.nb = nb_out;
		rc = filterset_enumerate_entries(fs,
					dump_config_filterset_add_entry,
					&ea);

		li = slist_get_next(li);
	}

	// filterset associations
	for (i=0 ; i<_MILTER_STATE_COUNT ; i++) {
		li = priv_data.state_filterset_list[i];
		while (li) {
			fs = slist_get_data(li);
			rc = snprintf (buffer,
				LINEBUFFER_SIZE,
				"pfm command milter stage attach name %s"
				" filterset %s\n",
				milter_state_name[i],
				fs->name);
			rc = pnetbuffer_add_string(nb_out, buffer);
			if (rc) return rc;
			li = slist_get_next(li);
		}
	}

	return 0;
}


